package org.ibex.jinetd;
import org.ibex.util.*;
import org.ibex.io.*;
import java.io.*;
import java.util.*;
import java.net.*;
import java.util.zip.*;

public class Port extends Loader {

    private final NetworkInterface iface;
    private final int port;
    private final Thread listener;

    public Port(String path, NetworkInterface iface, int port) {
        super(path);
        this.iface = iface;
        this.port = port;
        this.listener = new PortThread();
        listener.start();
    }

    public void changed(Watched w) {
        //Log.warn(this, "Port: noticed change in " + w);
        super.changed(w);
    }

    void dispatch(final Connection conn) throws Exception {
        getClassLoader();
        String[] list = list();
        for(int i=0; i<list.length; i++) {
            if (!list[i].endsWith(".jar")) continue;
            //Log.warn(this, "checking " + (this.path + File.separatorChar + list[i]));
            File f = new File(this.path + File.separatorChar + list[i]);
            ZipInputStream zis = new ZipInputStream(new FileInputStream(f));
            for(ZipEntry ze = zis.getNextEntry(); ze != null; ze = zis.getNextEntry()) {
                String name = ze.getName();
                if (name.endsWith(".class"))
                    dispatch(conn, name.substring(0, name.length() - ".class".length()).replace('/', '.'));
            }
        }
        check(conn, new File(getAbsolutePath() + File.separatorChar + "BIN"));
    }

    void check(Connection conn, File f) throws Exception {
        //Log.warn(this, "check(" + f.getAbsolutePath() + ")");
        if (!f.exists()) return;
        if (!f.isDirectory()) {
            if (!f.getAbsolutePath().endsWith(".class")) return;
            String name = f.getAbsolutePath().substring(getAbsolutePath().length() + 5);
            name = name.substring(0, name.length() - ".class".length()).replace(File.separatorChar, '.');
            dispatch(conn, name);
        } else {
            String[] list = f.list();
            for(int i=0; i<list.length; i++) check(conn, new File(f.getAbsolutePath() + File.separatorChar + list[i]));
        }
    }
    void dispatch(final Connection conn, String name) throws Exception {
        //Log.info(this, "attempting class " + name);
        try {
            Class c = getClassLoader().loadClass(name);
            if (c != null) {
                if (Listener.class.isAssignableFrom(c)) {
                    Log.error(this, "dispatching connection on port " + port + " to " +
                              c.getName());
                    final Listener l = (Listener)c.newInstance();
                    new Thread() { public void run() {
                        Log.clearnotes();
                        try {
                            l.accept(conn);
                        } catch (Exception e) {
                            Log.error(l.getClass(), "Listener threw exception");
                            Log.error(l.getClass(), e);
                        } finally {
                            conn.close();
                        }
                    } }.start();
                    return;
                }
            }
        } catch (Exception e) {
            Log.error(this, e);
        }
    }

    private class PortThread extends Thread {
        public void run() {
            try {
                Log.warn(this, "Now listening on interface " + iface + ", port " + port);
                ServerSocket ss = new ServerSocket(port);
                for(Socket s = ss.accept(); ; s = ss.accept()) try {
                    Log.warn(this, "accepted connection on port " + port);
                    dispatch(new Connection(s, "megacz.com"));
                    Log.warn(this, "done searching for service on port " + port);
                } catch (Exception e) { Log.warn(Port.class, e); }
            } catch (Exception e) { Log.error(Port.class, e);
            } catch (Throwable t) {
                Log.error(this, "serious error, aborting VM");
                Log.error(this, t);
                Root.reboot();
            }
        }
    }
}
