package org.ibex.jinetd;
import org.ibex.util.*;
import java.io.*;
import java.util.*;
import java.net.*;

public class Root extends Loader {

    private final Host host;
    private final Watched port;

    public Root(String path) {
        super(path);
        host = new Host(path + File.separatorChar + "host", null);
        port = new PortDir(path + File.separatorChar + "port");
    }

    public Watched slash(String part) {
        if (part.equals("host")) return host;
        if (part.equals("port")) return port;
        if (part.equals("LIB")) return super.slash(part);
        return null;
    }

    public static void reboot() {
        Log.flush();
        System.exit(0);
    }

    public void changed(Watched w) {
        if (w.part.equals("host")) {
            Log.warn(this, "/host changed");
        } else if (w.part.equals("port")) {
            Log.warn(this, "/port changed");
        } else if (w.getAbsolutePath().startsWith("/jinetd/LIB/")) {
            if (w.lastModifiedAtLastScan != -1) {
                Log.error(this, "jinetd upgraded; bouncing the JVM....");
                reboot();
            }
        } else {
            Log.info(this, "unknown directory " + w.part + " changed");
        }
    }

    private static class PortDir extends Watched {
        public PortDir(String path) { super(path); }
        public Watched slash(String part) {
            String ipaddr  = part.indexOf(':') == -1 ? null : part.substring(0, part.indexOf(':'));
            String portnum = part.indexOf(':') == -1 ? part : part.substring(part.indexOf(':') + 1);
            try {
                return new Port(this.path + File.separatorChar + part,
                                ipaddr == null ? null : NetworkInterface.getByInetAddress(InetAddress.getByName(ipaddr)),
                                portnum.equals("*") ? 0 : Integer.parseInt(portnum));
            } catch (UnknownHostException e) {  Log.warn(this, "can't resolve host for port directory: " + part);
            } catch (NumberFormatException e) { Log.warn(this, "invalid port directory: " + part);
            } catch (Exception e) {             Log.warn(this, "error instantiating Port: " + part);
            }
            return null;
        }
    }

}
